/*
 * 	BOOT LOADER v3.11 Copyright (C) 2006 HI-TECH Software
 *	This software is freely distributable and may be used
 *	for any purpose. No warranty of any kind is provided
 *	and all use is entirely at your own risk.
 *
 *	The boot loader is suitable for use with the following
 *	PICmicro controllers:
 *
 *	16F870, 16F871, 16F873, 16F873A, 16F874,
 *	16F874A, 16F876, 16F876A, 16F877, 16F877A
 *
 *	Refer to the accompanying bootldr.txt file for details.
 *	
 *	BOOTLDR.C
 */

#include <pic.h>
#include "bootldr.h"

/* Unprotect memory and disable the watchdog timer */
#ifdef MPLAB_ICD
__CONFIG(UNPROTECT & WDTDIS & DEBUGEN & LVPDIS);
#else
__CONFIG(UNPROTECT & WDTDIS);
#endif

// Macros to stop/start data TX at other end
#ifdef VERBOSE
 #define TX(mode) putch(mode)

// TX_STR is a cheaper version of puts()
#define TX_STR(str) 	EEADRH=(((unsigned int)(str))>>8)&0xff; \
			asm("movlw (_"___mkstr(str)")&0xff"); \
			asm("movwf _EEADR"); \
			tx_str();

#else
 #define TX(mode) TXREG=(mode)
#endif

#define _FWMASK (_FLASH_WRITE_SIZE - 1)

// "persistent" variables will not require initiazation routines
// at startup and therefore code size will be smaller.
persistent unsigned char cksum, rectype, bytecount;	// hex record attributes
persistent unsigned char databuff[16];	// buffer for hex data bytes
persistent unsigned char ovrbuff[32];	// buffer for incoming bytes after XOFF
persistent unsigned char overrun, ovrptr;
persistent BANKX unsigned char words2write;

persistent bit data_on;

persistent BANKX bit	isStartup;
persistent BANKX char tmpadrh;
#ifndef _BANK_ALIGNED
persistent BANKX persistent char tmpadr;
#endif
persistent BANKX persistent char ix;
#ifdef _WRITE_4_WORDS
persistent BANKX persistent char prewrite;
#endif



// if "VERBOSE" is defined, additional feedback is given by the below routines
#ifdef VERBOSE
const unsigned char START_MSG[]="HI-TECH Software (C)2006\n";
const unsigned char DOWNLOAD_MSG[]="\rSend ";	// Contains "end" (Ok message)

/* Output a byte to the serial port */
void 
putch(unsigned char byte)
{ 
	while(!TXIF);	// set when register is empty
		TXREG = byte;	// output one byte
}

// Output a const string via the serial port, EEADR already loaded 
// saved loading and unloading of parameter.
// I guarentee that this string is in ROM.
void 
tx_str(void){
	do{
		EEPGD=1;	// select flash for reading
		RD=1;
		NOP();
		NOP();
		putch(EEDATA);
		if(++EEADR==0)
			EEADRH++;
	}while(EEDATA);
}
#endif

// Recieve a character from the USART.
// in VERBOSE=2 mode this character is also echoed back.
unsigned char 
getch(void){
#if VERBOSE == 2
	static persistent unsigned char c;

	if(overrun){	// draw data from buffer or USART
		overrun--;
		c = ovrbuff[ovrptr++];	// return a byte from overflow buffer
	}else{
		if(! data_on){
			data_on=1;
			TX(XON);
		}
		while(!RCIF);	// reading from USART - wait for incoming
		c = RCREG;
	}
	putch(c);
	return c;
#else
	if(overrun){	// draw data from buffer or USART
		overrun--;
		return ovrbuff[ovrptr++];	// return a byte from overflow buffer
	}
	if(! data_on){ // need data and the buffer has run out, get from USART
		data_on=1;
		TX(XON);
	}
	while(!RCIF);	// reading from USART - wait for incoming
	return RCREG;
#endif
}

/* Get a 4 bit HEX number from the serial port */
unsigned char 
getXnibble(void)
{
    static persistent unsigned char tmp;
	tmp = getch();
	tmp -= '0';
	if (tmp>9)
		tmp -= ('A' - '0' - 10);
	return tmp;
}

/* Get an 8 bit HEX number from the serial port */
unsigned char 
getXbyte(void)
{
	static persistent unsigned char byte;
	byte=getXnibble();	// get first nibble
	byte=(byte<<4)|(byte>>4);	// nibble swap: make room for next nibble
	byte|=getXnibble();	// get second nibble
	cksum+=byte;		// compute checksum
	return (byte);
}

/* check the checksum */
void 
checksum(void)
{
	getXbyte();
	if(cksum!=0){	// if checksum does not add to zero, bad check, reset
#if VERBOSE
		TX('X');
#endif
		RESET();
	}
#if VERBOSE == 2
	putch('\r');	// echo each hex record on a new line
#endif
}

void 
main(void)
{
	INIT_COMMS();	// sets up the serial port for communication

	if(RCREG);  //flush receive register
	overrun=0;	

/* Bootloader waits for a specified time. If the serial port gets no response in */
/* this time, execution of existing program begins. */
#ifdef VERBOSE
	TX_STR(START_MSG);	// print a welcome message
#endif

	INTCON=0;	// disables interrupts.
	T1CON=0x35;	// pause during countdown
	bytecount=BOOT_TIMEOUT;
	TX(XON);	// we're open for business
	data_on = 1;
	do{
#ifdef VERBOSE	// display a countdown for user response 
		TX_STR(DOWNLOAD_MSG);		
		putch('0'+bytecount);	// "Send 9,8,7,6..."
#else
		while(!TXIF);	// wait for XON or \r to comlete
		TXREG=('0'+bytecount);	// 9,8,7,6...
#endif
		while(!TMR1IF); //wait...
		TMR1IF=0;	//clear the flag
		if(RCIF){
			TXREG=':';	// prompt indicates loader is ready for hex file
			break;
		}
#ifndef VERBOSE	// This is built into DOWNLOAD_MSG in verbose mode
		TXREG='\r';				// carriage return
#endif
	}while(--bytecount);
	T1CON=0x00;	// disable the timer again

	if (!RCIF)	//did it timeout without recieving anything?
	{	// no hex file to download, resume normal program
		#asm
run_program:	;Return here if programming succeeds
		ljmp BOOT_START
		#endasm	
	}

	/* receive a hex file via the serial port and write it to program memory */
	for(;;){		// loop until end of file
		// A union to access a word in its byte parts
		typedef union {
			unsigned int word;
			unsigned char byte[2];
		} wordbyte;

		static persistent unsigned char inx;
		static persistent BANKX wordbyte addr;

		while(getch() != ':');		// wait for start of hex record

#if VERBOSE == 1
		putch('.');	// print a dot for every incoming record
#endif

		words2write = cksum = bytecount = getXbyte();	// get the byte count and reset the checksum
		words2write >>= 1;	// convert byte count to word count
		addr.byte[1] = getXbyte();	// get the high byte of address
		addr.byte[0] = getXbyte();	// get the low byte of the address
		addr.word >>= 1;		// convert byte address to word address

		EEADR = addr.byte[0];
		EEADRH = addr.byte[1];

		if(addr.byte[1] == 0x21)	// unless this case,
			EEPGD = 0;	// when EEPROM should be selected
		else{
			EEPGD = 1;		// else destination is flash memory
		// ignore code that overlaps loader or data such as CONFIG/IDLOC
#ifdef _BANK_ALIGNED
			if( addr.byte[1] >= (BOOT_START>>8) ){
#else
			if( addr.word >= (BOOT_START) ){
#endif
				continue;
			}
		}

		rectype = getXbyte();	// get the record type
		inx=0;
		while(bytecount--){
			databuff[inx++]=getXbyte();	// read all bytes in this record
		}
		checksum();	// test the checksum of this record
		data_on=0;
		TX(XOFF);	// turn off the incoming stream before going to program
		ovrptr=0;	// reset the pointer into the over flow buffer

		T1CON=0x05;	// Sit and catch any late TX before proceeding
		do{
			if(RCIF)
				ovrbuff[overrun++]=RCREG;	// catch any late transmissions after stop request
		}while(!TMR1IF);
		TMR1IF=0;	// waited long enough - continue
		T1CON=0x00;	// disable the timer again

		if(rectype==1){		// if this was an END OF FILE record: prepare to run new program
#ifdef VERBOSE
#if VERBOSE == 2
			TXREG='\n';
#endif
			TX_STR(DOWNLOAD_MSG+2);	// message = "end"
#else
			TXREG=')';
#endif
#asm
		goto run_program
#endasm
		}

		// else ... INHX8M data record
		ix=0;
		isStartup=0;

		// Initiate a write to memory - EEADR has already been loaded
		if (((tmpadrh=addr.byte[1])|(addr.byte[0] & 0xFC)) == 0){	// is this address < 0004h?
			EEADRH = BOOT_START >> 8;	// write this into the bootloader area
#ifndef _BANK_ALIGNED
			tmpadr = EEADR;
			EEADR = ((BOOT_START & 0xFF));
#endif
			isStartup=1;	// Yes - this is the reset vector code of new app
		}

#ifdef _WRITE_4_WORDS
		prewrite = EEADR & _FWMASK;
		EEADR &= ~_FWMASK;	// point to start of flash-write boundary

		do{
			if(prewrite){
				prewrite--;
				RD=1;
				NOP();
				NOP();
			}else{
				EEDATA=databuff[ix++];
				EEDATH=databuff[ix++];
				if(words2write==0){
					prewrite=((~EEADR)&_FWMASK);
				}
			}
#else
		do{
			EEDATA=databuff[ix++];
			EEDATH=databuff[ix++];
#endif
			WREN=1;	// commit this word to flash
			EECON2=0x55;
			EECON2=0xAA;
			WR=1;			//initiate the write
			NOP();
			NOP();
			WREN=0;

			// Revert to real addresses if new reset vector was just written
#ifdef _BANK_ALIGNED
			if(isStartup && (EEADR==3)){
#else
			if(isStartup && ((EEADR&3)==3)){
#endif
				isStartup = 0;
				EEADRH = tmpadrh;	//swap back the address
#ifndef _BANK_ALIGNED
				EEADR = (tmpadr|3);
#endif
			}
			if(++EEADR == 0)
				EEADRH++;
#ifdef _WRITE_4_WORDS
		}while((prewrite)||words2write--);
#else
		}while(words2write--);
#endif
	}
}


#asm

	; Reset vector redirection

	; some equates for accessing the program counter 
	PCL	equ	02h		
	PCLATH	equ	0Ah

	psect	redirect,abs,class=CODE,delta=2	;absolute psect so it goes
						;to address zero


	;The boot loader code has been written so that it doesn't need any 
	;variable initialization runtime startup code.
	;Because of this, we can jump directly into the main() function. 
	global	_main		
	movlw	_main >> 8	;A "long" jump to main - this will cope with
	movwf	PCLATH		;main being positioned anywhere
	movlw	_main & 0xFF
	movwf	PCL

	; install a default reset image here. This will get overwritten by new program
	psect	init,class=CODE,delta=2
	movlw	_main >> 8	; default just returns us to main
	movwf	PCLATH		; must be exactly 4 words in length
	movlw	_main & 0xFF
	movwf	PCL
#endasm	

