#define DEBUG 1
/* #define SIMULATE 1 */

/*--------------------------------------------------------------
 *
 * Name:        SWEEP.C
 *
 * Purpose:     Send a frequency sweep command to the DDS.
 *
 *
 * Inputs:      Command line.   See usage() for information.
 *
 * Outputs:     RS-232 port.
 *
 * History:     Original, John Langner, WB2OSZ.
 *
 *--------------------------------------------------------------*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <bios.h>


#include "sg_cmd.h"

void long5_from_double (byte_t b[DDS_PREC], double f);

void long5_add (byte_t *psum, byte_t *padd);
void long5_negate (byte_t *pnum);

void simulate (struct sweep_cmd_s cmd);

void usage();



int main (int argc, char *argv[])
{
        double cycle = pow (2.0, DDS_PREC * 8);
        FILE *fp;

/*
 * From the command line.
 */
        int com_port_num;               /* For communicating with device. */
        double center_freq;             /* Center of sweep range. */
        double sweep_width;             /* Total width of sweep range. */
        int num_markers;                /* Z output calibration points. */ 
        int marker_spacing;             /* X output steps between them. */

/*
 * Calcuated items.
 */
        double min_freq;                /* Bottom end of sweep range. */
        double freq_step;               /* Delta F for each X step. */
        int n;
        struct sweep_cmd_s cmd;         /* Command parameters. */

        char port_name[10];


/*
 * Get the items from the command line.
 */
        com_port_num = 1;
        num_markers = 11;       /* Default. */
        marker_spacing = 0;     /* Default depends on num_markers, later. */

        n = 1;

        if (n >= argc) {
            usage();
        }

        if (strnicmp(argv[n], "/com=", 5) == 0) {
            com_port_num = atoi(argv[n]+5);
            printf ("Using COM port %d.\n", com_port_num);
            n++;
        }


        if (n >= argc) {
            fprintf (stderr, "Center frequency is missing.\n\n");
            usage();
        }

        center_freq = atof(argv[n]);
        n++;


        if (n >= argc) {
            fprintf (stderr, "Sweep width is missing.\n\n");
            usage();
        }

        sweep_width = atof(argv[n]);
        n++;


        if (n < argc) {
            num_markers = atoi(argv[n]);
            n++;
        }

        if (n < argc) {
            marker_spacing = atoi(argv[n]);
            n++;
        }


        if (n < argc) {
            fprintf (stderr, "Too many command line arguments.\n\n");
            usage ();
        }


/*
 * A little sanity checking.
 */
        if (center_freq <= 0) {
            fprintf (stderr, "Center frequency must be > zero.\n");
            exit (1);
        }

        if (sweep_width <= 0) {
            fprintf (stderr, "Sweep width must be > zero.\n");
            exit (1);
        }

        if (sweep_width > 2 * center_freq) {
            fprintf (stderr, "Warning: Sweep width reduced to twice center frequency.\n");
            sweep_width = 0.5 * center_freq;
        }

        if (num_markers < 2) {
            fprintf (stderr, "Warning: Number of markers must be at least 2.\n");
            num_markers = 2;
        }

        if (num_markers > 255) {
            /* Far larger than anything sensible. */
            fprintf (stderr, "Warning: Number of markers reduced to 255.\n");
            num_markers = 255;
        }

        if (marker_spacing == 0) {
            marker_spacing = 255 / (num_markers - 1);
        }

        if (marker_spacing < 2) {
            fprintf (stderr, "Warning: Marker spacing must be at least 2.\n");
            marker_spacing = 2;
        }

        if (marker_spacing > 255) {
            fprintf (stderr, "Warning: Marker spacing reduced to 255.\n");
            marker_spacing = 255;
        }

/*
 * There cases where we might want a very wide/slow sweep and
 * don't care about the X output.  Warn and continue.
 */
        if ((num_markers - 1) * marker_spacing + 1 > 256L) {
            fprintf (stderr, "Warning: X output will wrap around.\n");
        }

/*
 * Calculate frequency increment value.
 */
        freq_step = sweep_width / ((num_markers - 1L) * marker_spacing);
        min_freq = center_freq - 0.5 * sweep_width;

#ifdef DEBUG
        printf ("min freq = %.1f\n", min_freq);
        printf ("freq step = %.3f\n", freq_step);
#endif


/*
 * Fill in the command parameters.
 * +127.5 is for roundoff when only top 32 bits used.
 */
        long5_from_double (cmd.lower_f, min_freq * cycle / DDS_CLOCK + 127.5);
        long5_from_double (cmd.f_step, freq_step * cycle / DDS_CLOCK);
        cmd.n_marks = num_markers;
        cmd.spacing = marker_spacing;

#if DEBUG
        printf ("cmd: %c\n", SWEEP_CMD);

        printf ("  lower_f: %02x %02x %02x %02x %02x\n",
                cmd.lower_f[4],
                cmd.lower_f[3],
                cmd.lower_f[2],
                cmd.lower_f[1],
                cmd.lower_f[0]);

        printf ("  f_step: %02x %02x %02x %02x %02x\n",
                cmd.f_step[4],
                cmd.f_step[3],
                cmd.f_step[2],
                cmd.f_step[1],
                cmd.f_step[0]);

        printf ("  n_marks: %d\n", cmd.n_marks);
        printf ("  spacing: %d\n", cmd.spacing);
#endif

/*
 * Send it out serial port.
 * This is a one way message; no ack comes back.
 *
 * _bios_serialcom is for Borland C.
 * Might not work for other compilers.
 */


#ifdef SIMULATE

        simulate (cmd);
#else

        _bios_serialcom (_COM_INIT,
                com_port_num - 1,               /* 0 for COM1. */
                _COM_CHR8 | _COM_STOP1 | _COM_NOPARITY | _COM_300);

        sprintf (port_name, "COM%d", com_port_num);

        fp = fopen (port_name, "wb");
        if (fp == NULL) {
            fprintf (stderr, "ERROR: Couldn't open device %s.\n", port_name);
            exit (1);
        }

        fputc (SWEEP_CMD, fp);

        fwrite (&cmd, sizeof(cmd), 1, fp);

        fclose (fp);

#endif

        return (0);     /* Quiet compiler warning. */
}



/*-----------------------------------------------------------------
 *
 * Name:        simulate
 *
 * Purpose:     Simulate expected result before trying in PIC.
 *
 *----------------------------------------------------------------*/

void simulate (struct sweep_cmd_s cmd)
{
        byte_t x, z, m, s;      /* In real target too. */
        byte_t x_step;

        double actual_f;        /* For simulation only. */
        int j;

        printf ("\n");
        printf ("  X   Xstep   Z        F         \n");
        printf (" ---   ---   ---   -----------   \n");

        x = 0;
        x_step = 1;

        for (j=0; j<5; j++) {           /* Forever on real target. */

            m = cmd.n_marks - 1;
            do {

                s = cmd.spacing;
                do {

                    long5_add (cmd.lower_f, cmd.f_step);
                    x += x_step;

                    if (s == 1)
                        z = 1;
                    else
                        z = 0;

                    actual_f = (cmd.lower_f[4] * 256. * 256. * 256. +
                                cmd.lower_f[3] * 256. * 256. +
                                cmd.lower_f[2] * 256. +
                                cmd.lower_f[1] ) * DDS_CLOCK /
                                        pow (2.0, 32);
                  
                    printf (" %3d  %3d     %d  %11.2f\n",
                                x, x_step, z, actual_f); 

                } while (--s);

            } while (--m);

            long5_negate (cmd.f_step);
            x_step = ( - x_step);
        }

} /* end simulate */



/*-----------------------------------------------------------------
 *
 * Name:        long5_from_double
 *
 * Purpose:     Convert floating point value to 5 byte integer.
 *
 *----------------------------------------------------------------*/

void long5_from_double (byte_t b[DDS_PREC], double f)
{
        int n;

        f = floor(f + 0.5);
        for (n=0; n<DDS_PREC; n++) {

            b[n] = fmod(f, 256.);
            f = floor(f / 256.);
        }
}


/*------------------------------------------------------------------
 *
 * Name:        long5_add
 *
 * Purpose:     Add two five byte numbers. 
 *
 * Inputs:      psum, padd      - Add these two numbers.
 *                                Actually pointers to LSB.
 *
 * Outputs:     psum            - Their sum goes here.
 *
 * Description: It would have been a lot easier to use 4 byte
 *              numbers for all the calculations because the compiler
 *              has a built in "long" integer type.
 *              However, there would be excessive accumulation of
 *              round off errors when doing a lot of adding, especially
 *              for the sweep function.
 *
 *------------------------------------------------------------------*/

void long5_add (byte_t *psum, byte_t *padd)
{
        byte_t i;               /* Counter for looping. */
        byte_t kerry;           /* Carry from one byte to next. */

        int CARRY;              /* In status register on PIC. */

        kerry = 0;

        i = DDS_PREC;
        do {
/*
 * Apply any carry from preceding byte position.
 */
            CARRY = (*psum + kerry) > 255;

            *psum += kerry;

	    if (CARRY) {
		kerry = 1;	/* Will carry into next byte. */
	    }
	    else {
	        kerry = 0;	/* Didn't overflow. */
	    }
/*
 * Add bytes; keep track of any additional carry.
 */
 	    /* Compiler error: *psum += *padd */

            CARRY = (*psum + *padd) > 255;

            *psum = *psum + *padd;

            if (CARRY) {
                kerry++;
            }

            psum++;
            padd++;

        } while (--i); 
}



/*------------------------------------------------------------------
 *
 * Name:        long5_negate
 *
 * Purpose:     Negate a five byte numbers. 
 *
 * Inputs:      pnum            - Pointer to LSB of 5 bytes.
 *
 * Outputs:     pnum            - Result goes into same place.
 *
 *------------------------------------------------------------------*/

void long5_negate (byte_t *pnum)
{
        byte_t i;               /* Counter for looping. */
        byte_t kerry;           /* Carry from one byte to next. */
	
        int CARRY;              /* In status regiter on PIC. */

        kerry = 1;

        i = DDS_PREC;
        do {
            CARRY = ( ( ~ *pnum) & 0xff ) + kerry > 255;

            *pnum = ( ~ *pnum) + kerry;

	    if (CARRY) {
		kerry = 1;	/* Will carry into next byte. */
	    }
	    else {
	        kerry = 0;	/* Didn't overflow. */
	    }
	    
            pnum++;

        } while (--i);

}



/*-----------------------------------------------------------------
 *
 * Name:        usage
 *
 * Purpose:     Print brief usage message and exit.
 *
 *----------------------------------------------------------------*/

void usage ()
{
        fprintf (stderr, "usage: \n");
        fprintf (stderr, "        SWEEP [ /COM=n ] center width [ markers [ spacing ]]\n");
        fprintf (stderr, "where,\n");
        fprintf (stderr, "        n         is optional COM port number.  Default 1.\n");
        fprintf (stderr, "        center    is center frequency.\n");
        fprintf (stderr, "        width     is total frequency sweep width.\n");
        fprintf (stderr, "        markers   is number of Z output calibration markers.  Default 11.\n");
        fprintf (stderr, "        spacing   is number of X output steps per marker.\n");
        fprintf (stderr, "\n");

        exit (1);

} /* end usage */

/* end SWEEP.C */
